@@ -13,10 +13,11 @@ use notify::{RecommendedWatcher, RecursiveMode, Watcher};
 
 use calcit_runner::{
   builtins, call_stack, cli_args, codegen, codegen::emit_js::gen_stack, codegen::COMPILE_ERRORS_FILE, program, runner, snapshot, util,
+  ProgramEntries,
 };
 use im_ternary_tree::TernaryTreeList;
 
-pub struct ProgramSettings {
+pub struct CLIOptions {
   entry_path: PathBuf,
   emit_path: String,
   reload_libs: bool,
@@ -32,7 +33,7 @@ fn main() -> Result<(), String> {
   injection::inject_platform_apis();
 
   let cli_matches = cli_args::parse_cli();
-  let settings = ProgramSettings {
+  let cli_options = CLIOptions {
     // has default value
     entry_path: Path::new(cli_matches.value_of("input").unwrap()).to_owned(),
     emit_path: cli_matches.value_of("emit-path").or(Some("js-out")).unwrap().to_owned(),
@@ -59,7 +60,7 @@ fn main() -> Result<(), String> {
     }
     if let Some(cli_deps) = cli_matches.values_of("dep") {
       for module_path in cli_deps {
-        let module_data = calcit_runner::load_module(module_path, settings.entry_path.parent().unwrap())?;
+        let module_data = calcit_runner::load_module(module_path, cli_options.entry_path.parent().unwrap())?;
         for (k, v) in &module_data.files {
           snapshot.files.insert(k.to_owned(), v.to_owned());
         }
@@ -67,12 +68,12 @@ fn main() -> Result<(), String> {
     }
   } else {
     // load entry file
-    let content = fs::read_to_string(settings.entry_path.to_owned())
-      .unwrap_or_else(|_| panic!("expected Cirru snapshot: {:?}", settings.entry_path));
+    let content = fs::read_to_string(cli_options.entry_path.to_owned())
+      .unwrap_or_else(|_| panic!("expected Cirru snapshot: {:?}", cli_options.entry_path));
 
     let data = cirru_edn::parse(&content)?;
     // println!("reading: {}", content);
-    snapshot = snapshot::load_snapshot_data(data, settings.entry_path.to_str().unwrap())?;
+    snapshot = snapshot::load_snapshot_data(data, cli_options.entry_path.to_str().unwrap())?;
 
     // config in entry will overwrite default configs
     if let Some(entry) = cli_matches.value_of("entry") {
@@ -86,7 +87,7 @@ fn main() -> Result<(), String> {
 
     // attach modules
     for module_path in &snapshot.configs.modules {
-      let module_data = calcit_runner::load_module(module_path, settings.entry_path.parent().unwrap())?;
+      let module_data = calcit_runner::load_module(module_path, cli_options.entry_path.parent().unwrap())?;
       for (k, v) in &module_data.files {
         snapshot.files.insert(k.to_owned(), v.to_owned());
       }
@@ -94,6 +95,16 @@ fn main() -> Result<(), String> {
   }
   let init_fn = cli_matches.value_of("init-fn").or(Some(&snapshot.configs.init_fn)).unwrap();
   let reload_fn = cli_matches.value_of("reload-fn").or(Some(&snapshot.configs.reload_fn)).unwrap();
+  let (init_ns, init_def) = util::string::extract_ns_def(init_fn)?;
+  let (reload_ns, reload_def) = util::string::extract_ns_def(reload_fn)?;
+  let entries: ProgramEntries = ProgramEntries {
+    init_fn: init_fn.into(),
+    reload_fn: reload_fn.into(),
+    init_def: init_def.into(),
+    init_ns: init_ns.into(),
+    reload_ns: reload_ns.into(),
+    reload_def: reload_def.into(),
+  };
 
   // attach core
   for (k, v) in core_snapshot.files {
@@ -119,19 +130,20 @@ fn main() -> Result<(), String> {
   )
   .map_err(|e| e.msg)?;
 
-  let task = if settings.emit_js {
-    run_codegen(init_fn, reload_fn, &settings.emit_path, false)
-  } else if settings.emit_ir {
-    run_codegen(init_fn, reload_fn, &settings.emit_path, true)
+  let task = if cli_options.emit_js {
+    run_codegen(&entries, &cli_options.emit_path, false)
+  } else if cli_options.emit_ir {
+    run_codegen(&entries, &cli_options.emit_path, true)
   } else {
     let started_time = Instant::now();
 
-    let v = calcit_runner::run_program(init_fn, TernaryTreeList::Empty).map_err(|e| {
-      for w in e.warnings {
-        println!("{}", w);
-      }
-      e.msg
-    })?;
+    let v =
+      calcit_runner::run_program(entries.init_ns.to_owned(), entries.init_def.to_owned(), TernaryTreeList::Empty).map_err(|e| {
+        for w in e.warnings {
+          println!("{}", w);
+        }
+        e.msg
+      })?;
 
     let duration = Instant::now().duration_since(started_time);
     println!("took {}ms: {}", duration.as_micros() as f64 / 1000.0, v);
@@ -152,17 +164,16 @@ fn main() -> Result<(), String> {
 
   if !eval_once {
     runner::track::track_task_add();
-    let copied_init_fn = Arc::new(init_fn.to_owned());
-    let copied_reload_fn = Arc::new(reload_fn.to_owned());
     let copied_assets = Arc::new(assets_watch.map(|s| s.to_owned()).to_owned());
-    let copied_settings = Arc::new(settings);
-    std::thread::spawn(move || watch_files(copied_init_fn, copied_reload_fn, copied_settings, copied_assets));
+    let copied_settings = Arc::new(cli_options);
+    let copied_entries = Arc::new(entries);
+    std::thread::spawn(move || watch_files(copied_entries, copied_settings, copied_assets));
   }
   runner::track::exit_when_cleared();
   Ok(())
 }
 
-pub fn watch_files(init_fn: Arc<String>, reload_fn: Arc<String>, settings: Arc<ProgramSettings>, assets_watch: Arc<Option<String>>) {
+pub fn watch_files(entries: Arc<ProgramEntries>, settings: Arc<CLIOptions>, assets_watch: Arc<Option<String>>) {
   println!("\nRunner: in watch mode...\n");
   let (tx, rx) = channel();
   let mut watcher: RecommendedWatcher = Watcher::new(tx, Duration::from_millis(200)).unwrap();
@@ -192,7 +203,7 @@ pub fn watch_files(init_fn: Arc<String>, reload_fn: Arc<String>, settings: Arc<P
               println!("failed re-compiling, got empty inc file");
               continue;
             }
-            if let Err(e) = recall_program(&content, &init_fn, &reload_fn, &settings) {
+            if let Err(e) = recall_program(&content, &entries, &settings) {
               println!("error: {}", e);
             };
           }
@@ -208,7 +219,7 @@ pub fn watch_files(init_fn: Arc<String>, reload_fn: Arc<String>, settings: Arc<P
 
 // overwrite previous state
 
-fn recall_program(content: &str, init_fn: &str, reload_fn: &str, settings: &ProgramSettings) -> Result<(), String> {
+fn recall_program(content: &str, entries: &ProgramEntries, settings: &CLIOptions) -> Result<(), String> {
   println!("\n-------- file change --------\n");
 
   // Steps:
@@ -225,22 +236,41 @@ fn recall_program(content: &str, init_fn: &str, reload_fn: &str, settings: &Prog
   // println!("\nprogram code: {:?}", new_code);
 
   // clear data in evaled states
-  program::clear_all_program_evaled_defs(init_fn, reload_fn, settings.reload_libs)?;
+  program::clear_all_program_evaled_defs(entries.init_ns.to_owned(), entries.reload_ns.to_owned(), settings.reload_libs)?;
   builtins::meta::force_reset_gensym_index()?;
 
   let task = if settings.emit_js {
-    run_codegen(init_fn, reload_fn, &settings.emit_path, false)
+    run_codegen(entries, &settings.emit_path, false)
   } else if settings.emit_ir {
-    run_codegen(init_fn, reload_fn, &settings.emit_path, true)
+    run_codegen(entries, &settings.emit_path, true)
   } else {
     // run from `reload_fn` after reload
     let started_time = Instant::now();
-    let v = calcit_runner::run_program(reload_fn, TernaryTreeList::Empty).map_err(|e| {
-      for w in e.warnings {
-        println!("{}", w);
+    if runner::track::has_pending_than_watcher() {
+      // when there's services, make sure their code get preprocessed too
+      let check_warnings: &RefCell<Vec<String>> = &RefCell::new(vec![]);
+      println!("preprocesssing: {:?}", entries);
+      if let Err(e) = runner::preprocess::preprocess_ns_def(
+        entries.init_ns.to_owned(),
+        entries.init_def.to_owned(),
+        entries.init_def.to_owned(),
+        None,
+        check_warnings,
+        &rpds::List::new_sync(),
+      ) {
+        return Err(e.to_string());
       }
-      e.msg
-    })?;
+
+      let warnings = check_warnings.to_owned().into_inner();
+      throw_on_warnings(&warnings)?;
+    }
+    let v =
+      calcit_runner::run_program(entries.reload_ns.to_owned(), entries.reload_def.to_owned(), TernaryTreeList::Empty).map_err(|e| {
+        for w in e.warnings {
+          println!("{}", w);
+        }
+        e.msg
+      })?;
     let duration = Instant::now().duration_since(started_time);
     println!("took {}ms: {}", duration.as_micros() as f64 / 1000.0, v);
     Ok(())
@@ -256,12 +286,9 @@ fn recall_program(content: &str, init_fn: &str, reload_fn: &str, settings: &Prog
   Ok(())
 }
 
-fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -> Result<(), String> {
+fn run_codegen(entries: &ProgramEntries, emit_path: &str, ir_mode: bool) -> Result<(), String> {
   let started_time = Instant::now();
 
-  let (init_ns, init_def) = util::string::extract_ns_def(init_fn)?;
-  let (reload_ns, reload_def) = util::string::extract_ns_def(reload_fn)?;
-
   if ir_mode {
     builtins::effects::modify_cli_running_mode(builtins::effects::CliRunningMode::Ir)?;
   } else {
@@ -280,9 +307,9 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
 
   // preprocess to init
   match runner::preprocess::preprocess_ns_def(
-    init_ns.to_owned().into(),
-    init_def.to_owned().into(),
-    init_def.to_owned().into(),
+    entries.init_ns.to_owned(),
+    entries.init_def.to_owned(),
+    entries.init_def.to_owned(),
     None,
     check_warnings,
     &rpds::List::new_sync(),
@@ -305,9 +332,9 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
 
   // preprocess to reload
   match runner::preprocess::preprocess_ns_def(
-    reload_ns.into(),
-    reload_def.into(),
-    init_def.to_owned().into(),
+    entries.reload_ns.to_owned(),
+    entries.reload_def.to_owned(),
+    entries.init_def.to_owned(),
     None,
     check_warnings,
     &rpds::List::new_sync(),
@@ -319,21 +346,9 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
       return Err(failure.msg);
     }
   }
-  let warnings = check_warnings.to_owned().into_inner();
-  let mut content: String = String::from("");
-  if !warnings.is_empty() {
-    for message in &warnings {
-      println!("{}", message);
-      content = format!("{}\n{}", content, message);
-    }
 
-    let _ = fs::write(&js_file_path, format!("export default \"{}\";", content.trim().escape_default()));
-    return Err(format!(
-      "Found {} warnings, codegen blocked. errors in {}.js",
-      warnings.len(),
-      COMPILE_ERRORS_FILE,
-    ));
-  }
+  let warnings = check_warnings.to_owned().into_inner();
+  throw_on_js_warnings(&warnings, &js_file_path)?;
 
   // clear if there are no errors
   let no_error_code = String::from("export default null;");
@@ -342,7 +357,7 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
   }
 
   if ir_mode {
-    match codegen::gen_ir::emit_ir(init_fn, reload_fn, emit_path) {
+    match codegen::gen_ir::emit_ir(&*entries.init_fn, &*entries.reload_fn, emit_path) {
       Ok(_) => (),
       Err(failure) => {
         println!("\nfailed codegen, {}", failure);
@@ -352,7 +367,7 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
     }
   } else {
     // TODO entry ns
-    match codegen::emit_js::emit_js(&init_ns.to_owned(), emit_path) {
+    match codegen::emit_js::emit_js(&entries.init_ns, emit_path) {
       Ok(_) => (),
       Err(failure) => {
         println!("\nfailed codegen, {}", failure);
@@ -365,3 +380,36 @@ fn run_codegen(init_fn: &str, reload_fn: &str, emit_path: &str, ir_mode: bool) -
   println!("took {}ms", duration.as_micros() as f64 / 1000.0);
   Ok(())
 }
+
+fn throw_on_js_warnings(warnings: &[String], js_file_path: &Path) -> Result<(), String> {
+  if !warnings.is_empty() {
+    let mut content: String = String::from("");
+    for message in warnings {
+      println!("{}", message);
+      content = format!("{}\n{}", content, message);
+    }
+
+    let _ = fs::write(&js_file_path, format!("export default \"{}\";", content.trim().escape_default()));
+    return Err(format!(
+      "Found {} warnings, codegen blocked. errors in {}.js",
+      warnings.len(),
+      COMPILE_ERRORS_FILE,
+    ));
+  } else {
+    Ok(())
+  }
+}
+
+fn throw_on_warnings(warnings: &[String]) -> Result<(), String> {
+  if !warnings.is_empty() {
+    let mut content: String = String::from("");
+    for message in warnings {
+      println!("{}", message);
+      content = format!("{}\n{}", content, message);
+    }
+
+    return Err(format!("Found {} warnings, runner blocked.", warnings.len(),));
+  } else {
+    Ok(())
+  }
+}
